#/********************************************************************
# Cayenue/cayenue/panels/cameras/systemtab.py 
#
# Copyright (c) 2023  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

from PyQt6.QtWidgets import QGridLayout, QWidget, QPushButton, QGroupBox, \
    QMessageBox, QRadioButton, QComboBox, QLabel, QDialog, QDialogButtonBox, \
    QCheckBox, QLineEdit
from PyQt6.QtCore import Qt, QTimer
from datetime import datetime, timedelta
import webbrowser
import time
from cayenue.enums import ProxyType

class TimeDialog(QDialog):
    def __init__(self, mw):
        super().__init__(mw)
        self.mw = mw
        self.onvif_data = None
        self.timer = QTimer()
        self.timer.setInterval(1000)
        self.timer.timeout.connect(self.timeout)

        self.setWindowTitle("Time Settings")
        self.setMinimumWidth(640)

        self.lblComputerDate = QLabel()
        self.lblComputerTime = QLabel()
        self.lblComputerDST = QLabel()
        self.lblComputerTZ = QLabel()
        self.lblComputerUTCDiff = QLabel()
        self.updateComputerTime()        
        grpComputer = QGroupBox("Computer Time")
        lytComputer = QGridLayout(grpComputer)
        lytComputer.addWidget(QLabel("Date"),          0, 0, 1, 1)
        lytComputer.addWidget(self.lblComputerDate,    0, 1, 1, 1)
        lytComputer.addWidget(QLabel("Time"),          1, 0, 1, 1)
        lytComputer.addWidget(self.lblComputerTime,    1, 1, 1, 1)
        lytComputer.addWidget(QLabel("DST"),           2, 0, 1, 1)
        lytComputer.addWidget(self.lblComputerDST,     2, 1, 1, 1)
        lytComputer.addWidget(QLabel("Time Zone"),     3, 0, 1, 1)
        lytComputer.addWidget(self.lblComputerTZ,      3, 1, 1, 1)
        lytComputer.addWidget(QLabel("UTC Offset"),    4, 0, 1, 1)
        lytComputer.addWidget(self.lblComputerUTCDiff, 4, 1, 1, 1)

        self.lblCameraDate = QLabel()
        self.lblCameraTime = QLabel()
        self.lblCameraDST = QLabel()
        self.lblCameraTZ = QLabel()
        self.lblCameraUTCDiff = QLabel()
        self.lblCameraDiffLbl = QLabel("Diff (sec)")
        self.updateCameraTime()        
        grpCamera = QGroupBox("Camera Time")
        lytCamera = QGridLayout(grpCamera)
        lytCamera.addWidget(QLabel("Date"),        0, 0, 1, 1)
        lytCamera.addWidget(self.lblCameraDate,    0, 1, 1, 1)
        lytCamera.addWidget(QLabel("Time"),        1, 0, 1, 1)
        lytCamera.addWidget(self.lblCameraTime,    1, 1, 1, 1)
        lytCamera.addWidget(QLabel("DST"),         2, 0, 1, 1)
        lytCamera.addWidget(self.lblCameraDST,     2, 1, 1, 1)
        lytCamera.addWidget(QLabel("Time Zone"),   3, 0, 1, 1)
        lytCamera.addWidget(self.lblCameraTZ,      3, 1, 1, 1)
        lytCamera.addWidget(self.lblCameraDiffLbl, 4, 0, 1, 1)
        lytCamera.addWidget(self.lblCameraUTCDiff, 4, 1, 1, 1)

        self.txtTimezone = QLineEdit()
        self.lblTimezone = QLabel("Time Zone")
        self.txtNTP = QLineEdit()
        self.lblNTP = QLabel("NTP Server")
        self.chkDST = QCheckBox("Daylight Savings Time")
        self.cmbDateTimeType = QComboBox()
        self.cmbDateTimeType.addItem("Manual")
        self.cmbDateTimeType.addItem("NTP")

        self.radNTP = QRadioButton("NTP")
        self.radNTP.toggled.connect(self.radNTPToggled)
        self.radManual = QRadioButton("Manual")
        self.radManual.toggled.connect(self.radManualToggled)
        self.radUTCasLocal = QRadioButton("UTC as Local")
        self.radUTCasLocal.toggled.connect(self.radUTCasLocalToggled)

        grpSettings = QGroupBox("Time Sync Method")
        lytSettings = QGridLayout(grpSettings)
        lytSettings.addWidget(self.radNTP,             0, 0, 1, 1, Qt.AlignmentFlag.AlignCenter)
        lytSettings.addWidget(self.radManual,          0, 1, 1, 1, Qt.AlignmentFlag.AlignCenter)
        lytSettings.addWidget(self.radUTCasLocal,      0, 2, 1, 1, Qt.AlignmentFlag.AlignCenter)
        
        self.radFromDHCP = QRadioButton("From DHCP")
        self.radFromDHCP.toggled.connect(self.radFromDHCPToggled)
        self.radIPv4 = QRadioButton("IPv4 Address")
        self.radDNS = QRadioButton("Domain Name")
        
        self.grpNTPServer = QGroupBox("NTP Server")
        lytNTPServer = QGridLayout(self.grpNTPServer)
        lytNTPServer.addWidget(QLabel("NTP Server Address Type:"),   0, 1, 1, 1)
        lytNTPServer.addWidget(self.radFromDHCP,                     0, 2, 1, 1)
        lytNTPServer.addWidget(self.radIPv4,                         1, 2, 1, 1)
        lytNTPServer.addWidget(self.radDNS,                          2, 2, 1, 1)
        lytNTPServer.addWidget(self.lblNTP,                          4, 0, 1, 1, Qt.AlignmentFlag.AlignRight)
        lytNTPServer.addWidget(self.txtNTP,                          4, 1, 1, 3)

        buttonBox = QDialogButtonBox( \
            QDialogButtonBox.StandardButton.Ok | \
            QDialogButtonBox.StandardButton.Cancel)
        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)

        lytMain = QGridLayout(self)
        lytMain.addWidget(grpComputer,             0, 0, 1, 2)
        lytMain.addWidget(grpCamera,               0, 2, 1, 2)
        lytMain.addWidget(QLabel(),                1, 0, 1, 4)
        lytMain.addWidget(self.lblTimezone,        2, 0, 1, 1)
        lytMain.addWidget(self.txtTimezone,        2, 1, 1, 3)
        lytMain.addWidget(self.chkDST,             3, 0, 1, 4)
        lytMain.addWidget(grpSettings,             4, 1, 1, 2)
        lytMain.addWidget(self.grpNTPServer,       6, 0, 1, 4)
        lytMain.addWidget(buttonBox,               7, 0, 1, 4)

    def radFromDHCPToggled(self, arg):
        self.txtNTP.setEnabled(not arg)
        self.lblNTP.setEnabled(not arg)

    def radNTPToggled(self, arg):
        self.grpNTPServer.setEnabled(arg)
        if arg and self.onvif_data:
            self.onvif_data.setDateTimeType('N')

    def radManualToggled(self, arg):
        if arg and self.onvif_data:
            self.onvif_data.setDateTimeType('M')

    def radUTCasLocalToggled(self, arg):
        if arg and self.onvif_data:
            self.onvif_data.setDateTimeType('U')
        self.txtTimezone.setEnabled(not arg)
        self.lblTimezone.setEnabled(not arg)
        self.chkDST.setEnabled(not arg)

    def exec(self, onvif_data):
        self.onvif_data = onvif_data

        if onvif_data.datetimetype() == 'U':
            self.radUTCasLocal.setChecked(True)
            self.radNTPToggled(False)
            self.radManualToggled(False)
        if onvif_data.datetimetype() == 'M':
            self.radManual.setChecked(True)
            self.radNTPToggled(False)
        if onvif_data.datetimetype() == 'N':
            self.radNTP.setChecked(True)
            self.radManualToggled(False)

        if onvif_data.ntp_dhcp():
            self.radFromDHCP.setChecked(True)
        else:
            if onvif_data.ntp_type() == "IPv4":
                self.radIPv4.setChecked(True)
            if onvif_data.ntp_type() == "DNS":
                self.radDNS.setChecked(True)
        self.txtNTP.setText(onvif_data.ntp_addr())

        tz = onvif_data.timezone()
        self.lblCameraTZ.setText(tz)
        self.txtTimezone.setText(tz)
        self.lblCameraDST.setText(str(bool(onvif_data.dst())))
        self.txtNTP.setText(onvif_data.ntp_addr())
        self.chkDST.setChecked(onvif_data.dst())
        self.lblCameraUTCDiff.setText(str(onvif_data.time_offset()))

        self.updateComputerTime()
        self.updateCameraTime()

        self.timer.start()
        super().exec()

    def accept(self):
        self.onvif_data.setNTPDHCP(self.radFromDHCP.isChecked())
        if self.radIPv4.isChecked():
            self.onvif_data.setNTPType("IPv4")
        if self.radDNS.isChecked():
            self.onvif_data.setNTPType("DNS")
        self.onvif_data.setNTPAddr(self.txtNTP.text())
        self.onvif_data.setDST(self.chkDST.isChecked())
        self.onvif_data.setTimezone(self.txtTimezone.text())
        self.onvif_data.startUpdateTime()
        self.close()

    def reject(self):
        self.close()

    def closeEvent(self, e):
        self.timer.stop()

    def timeout(self):
        self.updateComputerTime()
        self.updateCameraTime()

    def updateComputerTime(self):
        now = datetime.now()
        self.lblComputerDate.setText(now.strftime("%b-%d-%Y"))
        self.lblComputerTime.setText(now.strftime('%H:%M:%S'))
        dst = bool(time.localtime().tm_isdst)
        self.lblComputerDST.setText(str(dst))
        local_now = now.astimezone()
        local_tz = local_now.tzinfo
        self.lblComputerTZ.setText(local_tz.tzname(local_now))
        self.lblComputerUTCDiff.setText(str(local_now)[-6:])

    def updateCameraTime(self):
        now = datetime.now()
        if self.onvif_data:
            offset = self.onvif_data.time_offset()
            if self.radUTCasLocal.isChecked():
                local_now = now.astimezone()
                utc_diff = local_now.utcoffset()
                seconds_diff = utc_diff.days * 24 * 60 * 60 + utc_diff.seconds
                offset -= seconds_diff
                self.lblCameraDiffLbl.setText("Diff (sec) *adj")
                self.lblCameraUTCDiff.setText(str(offset))
            else:
                self.lblCameraDiffLbl.setText("Diff (sec)")
                self.lblCameraUTCDiff.setText(str(self.onvif_data.time_offset()))

            delta = timedelta(seconds=offset)
            now += delta

        self.lblCameraDate.setText(now.strftime("%b-%d-%Y"))
        self.lblCameraTime.setText(now.strftime('%H:%M:%S'))

class SystemTabSettings():
    def __init__(self, mw, camera):
        self.camera = camera
        self.mw = mw

        self.record_enable = self.getRecordAlarmEnabled()
        self.record_always = self.getRecordAlways()
        self.record_alarm = self.getRecordOnAlarm()
        self.sound_alarm_enable = self.getSoundAlarmEnabled()
        self.sound_alarm_once = self.getSoundAlarmOnce()
        self.sound_alarm_loop = self.getSoundAlarmLoop()
        self.record_profile = self.getRecordProfile()
        self.remote_snapshot = self.getRemoteSnapshot()

    def managePlayers(self):
        record = False
        if self.record_enable:
            if self.record_always or (self.record_alarm and self.camera.isAlarming()):
                record = True
        if record:
            if profile := self.camera.getRecordProfile():
                if player := self.mw.pm.getPlayer(profile.uri()):
                    if not player.isRecording():
                        if filename := player.getOutputFilename():
                            player.toggleRecording(filename)
        else:
            players = self.mw.pm.getStreamPairPlayers(self.camera.uri())
            for player in players:
                if player.isRecording():
                    player.toggleRecording("")
                    self.mw.settingsPanel.storage.signals.updateDiskUsage.emit()
                
        self.mw.cameraPanel.syncGUI()

    def getRecordProfile(self):
        key = f'{self.camera.serial_number()}/RecordProfile'
        return int(self.mw.settings.value(key, 0))
    
    def setRecordProfile(self, ordinal):
        self.record_profile = ordinal
        key = f'{self.camera.serial_number()}/RecordProfile'
        self.mw.settings.setValue(key, ordinal)

    def getRecordAlarmEnabled(self):
        key = f'{self.camera.serial_number()}/RecordAlarmEnabled'
        return bool(int(self.mw.settings.value(key, 0)))
    
    def setRecordAlarmEnabled(self, state):
        self.record_enable = bool(state)
        key = f'{self.camera.serial_number()}/RecordAlarmEnabled'
        self.managePlayers()
        self.mw.settings.setValue(key, int(state))

    def getRecordAlways(self):
        key = f'{self.camera.serial_number()}/RecordAlways'
        return bool(int(self.mw.settings.value(key, 1)))
    
    def setRecordAlways(self, state):
        self.record_always = bool(state)
        key = f'{self.camera.serial_number()}/RecordAlways'
        self.mw.settings.setValue(key, int(state))
        self.managePlayers()

    def getRecordOnAlarm(self):
        key = f'{self.camera.serial_number()}/RecordOnAlarm'
        return bool(int(self.mw.settings.value(key, 0)))
    
    def setRecordOnAlarm(self, state):
        self.record_alarm = bool(state)
        key = f'{self.camera.serial_number()}/RecordOnAlarm'
        self.mw.settings.setValue(key, int(state))

    def getSoundAlarmEnabled(self):
        key = f'{self.camera.serial_number()}/SoundAlarmEnabled'
        return bool(int(self.mw.settings.value(key, 0)))

    def setSoundAlarmEnabled(self, state):
        self.sound_alarm_enable = bool(state)
        key = f'{self.camera.serial_number()}/SoundAlarmEnabled'
        self.mw.settings.setValue(key, int(state))

    def getSoundAlarmOnce(self):
        key = f'{self.camera.serial_number()}/SoundAlarmOnce'
        return bool(int(self.mw.settings.value(key, 0)))

    def setSoundAlarmOnce(self, state):
        self.sound_alarm_once = bool(state)
        key = f'{self.camera.serial_number()}/SoundAlarmOnce'
        self.mw.settings.setValue(key, int(state))

    def getSoundAlarmLoop(self):
        key = f'{self.camera.serial_number()}/SoundAlarmLoop'
        return bool(int(self.mw.settings.value(key, 1)))

    def setSoundAlarmLoop(self, state):
        self.sound_alarm_loop = bool(state)
        key = f'{self.camera.serial_number()}/SoundAlarmLoop'
        self.mw.settings.setValue(key, int(state))

    def getRemoteSnapshot(self):
        key = f'{self.camera.serial_number()}/RemoteSnapshot'
        return bool(int(self.mw.settings.value(key, 0)))

    def setRemoteSnapshot(self, state):
        self.remote_snapshot= bool(state)
        key = f'{self.camera.serial_number()}/RemoteSnapshot'
        self.mw.settings.setValue(key, int(state))

class SystemTab(QWidget):
    def __init__(self, cp):
        super().__init__()
        self.cp = cp

        self.dlgTimeDialog = TimeDialog(self.cp.mw)

        self.radRecordAlways = QRadioButton("Always")
        self.radRecordAlways.clicked.connect(self.radRecordAlwaysClicked)
        self.radRecordAlways.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.radRecordOnAlarm = QRadioButton("Alarms")
        self.radRecordOnAlarm.clicked.connect(self.radRecordOnAlarmClicked)
        self.radRecordOnAlarm.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.grpRecord = QGroupBox("Record")
        self.grpRecord.setCheckable(True)
        self.grpRecord.clicked.connect(self.grpRecordClicked)
        self.grpRecord.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        lytGroup = QGridLayout(self.grpRecord)
        lytGroup.addWidget(self.radRecordAlways,    0, 0, 1, 1)
        lytGroup.addWidget(self.radRecordOnAlarm,   1, 0, 1, 1)

        self.radSoundOnce = QRadioButton("Once")
        self.radSoundOnce.clicked.connect(self.radSoundOnceClicked)
        self.radSoundOnce.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.radSoundLoop = QRadioButton("Loop")
        self.radSoundLoop.clicked.connect(self.radSoundLoopClicked)
        self.radSoundLoop.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.grpSounds = QGroupBox("Sounds")
        self.grpSounds.setCheckable(True)
        self.grpSounds.clicked.connect(self.grpSoundsClicked)
        self.grpSounds.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        lytGroupSounds = QGridLayout(self.grpSounds)
        lytGroupSounds.addWidget(self.radSoundOnce,   0, 0, 1, 1)
        lytGroupSounds.addWidget(self.radSoundLoop,   1, 0, 1, 1)

        self.cmbRecordProfile = QComboBox()
        self.cmbRecordProfile.currentIndexChanged.connect(self.cmbRecordProfileChanged)
        self.cmbRecordProfile.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        #self.lblRecordResolution = QLabel("      ")
        self.lblRecordProfile = QLabel("  Record Profile   ")
        self.chkRecordAudio = QCheckBox("Record Audio")
        self.chkRecordAudio.stateChanged.connect(self.chkRecordAudioClicked)
        self.chkRecordAudio.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.chkRemoteSnapshot = QCheckBox("Remote Snapshot Image")
        self.chkRemoteSnapshot.stateChanged.connect(self.chkRemoteSnapshotChecked)
        self.chkRemoteSnapshot.setFocusPolicy(Qt.FocusPolicy.NoFocus)

        pnlRecordProfile = QWidget()
        lytRecordProfile = QGridLayout(pnlRecordProfile)
        lytRecordProfile.addWidget(self.lblRecordProfile,     0, 0, 1, 1)
        lytRecordProfile.addWidget(self.cmbRecordProfile,     0, 1, 1, 1)
        lytRecordProfile.addWidget(QLabel("               "), 0, 2, 1, 1)
        #lytRecordProfile.addWidget(self.lblRecordResolution,  1, 0, 1, 1)
        lytRecordProfile.addWidget(self.chkRecordAudio,       1, 0, 1, 1)
        lytRecordProfile.addWidget(self.chkRemoteSnapshot,       1, 1, 1, 2)
        lytRecordProfile.setColumnStretch(1, 5)
        lytRecordProfile.setContentsMargins(0, 0, 0, 0)

        self.btnReboot = QPushButton("Reboot")
        self.btnReboot.clicked.connect(self.btnRebootClicked)
        self.btnReboot.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.btnSyncTime = QPushButton("Sync Time")
        self.btnSyncTime.clicked.connect(self.btnSyncTimeClicked)
        self.btnSyncTime.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.btnBrowser = QPushButton("Browser")
        self.btnBrowser.clicked.connect(self.btnBrowserClicked)
        self.btnBrowser.setFocusPolicy(Qt.FocusPolicy.NoFocus)

        pnlButton = QWidget()
        lytButton = QGridLayout(pnlButton)
        lytButton.addWidget(self.btnReboot,     0, 0, 1, 1)
        lytButton.addWidget(self.btnSyncTime,   1, 0, 1, 1)
        lytButton.addWidget(self.btnBrowser,    2, 0, 1, 1)
        lytButton.setContentsMargins(6, 2, 6, 2)

        lytMain = QGridLayout(self)
        lytMain.addWidget(self.grpRecord,         0, 0, 1, 1)
        lytMain.addWidget(self.grpSounds,         0, 1, 1, 1)
        lytMain.addWidget(pnlRecordProfile,       1, 0, 1, 3)
        lytMain.addWidget(pnlButton,              0, 2, 1, 1)

    def grpRecordClicked(self, state):
        if camera := self.cp.getCurrentCamera():
            camera.systemTabSettings.setRecordAlarmEnabled(state)
            self.cp.mw.enable_disk_mgmt = self.cp.mw.pm.recording()

    def radRecordAlwaysClicked(self, state):
        if camera := self.cp.getCurrentCamera():
            camera.systemTabSettings.setRecordAlways(state)
            camera.systemTabSettings.setRecordOnAlarm(not state)

    def radRecordOnAlarmClicked(self, state):
        if camera := self.cp.getCurrentCamera():
            camera.systemTabSettings.setRecordOnAlarm(state)
            camera.systemTabSettings.setRecordAlways(not state)

    def grpSoundsClicked(self, state):
        if camera := self.cp.getCurrentCamera():
            camera.systemTabSettings.setSoundAlarmEnabled(state)

    def radSoundOnceClicked(self, state):
        if camera := self.cp.getCurrentCamera():
            camera.systemTabSettings.setSoundAlarmOnce(state)
            camera.systemTabSettings.setSoundAlarmLoop(not state)

    def radSoundLoopClicked(self, state):
        if camera := self.cp.getCurrentCamera():
            camera.systemTabSettings.setSoundAlarmLoop(state)
            camera.systemTabSettings.setSoundAlarmOnce(not state)

    def cmbRecordProfileChanged(self, index):
        if camera := self.cp.getCurrentCamera():
            players = self.cp.mw.pm.getStreamPairPlayers(camera.uri())
            camera.systemTabSettings.setRecordProfile(index)
            if len(players):
                for player in players:
                    self.cp.mw.pm.playerShutdownWait(player.uri)
                self.cp.onItemDoubleClicked(camera)
        self.syncGUI()

    def chkRecordAudioClicked(self, state):
        if camera := self.cp.getCurrentCamera():
            if profile := camera.getRecordProfile():
                profile.setDisableAudio(not state)
                if player := self.cp.mw.pm.getPlayer(profile.uri()):
                    if player.hasAudio():
                        if player.disable_audio == bool(state):
                            player.disable_audio = bool(not state)
                            self.cp.mw.pm.playerShutdownWait(player.uri)
                            self.cp.mw.playMedia(player.uri)

    def chkRemoteSnapshotChecked(self, state):
        if camera := self.cp.getCurrentCamera():
            camera.systemTabSettings.setRemoteSnapshot(state)

    def fill(self, onvif_data):
        self.cmbRecordProfile.currentIndexChanged.disconnect()
        self.cmbRecordProfile.clear()
        for profile in onvif_data.profiles:
            self.cmbRecordProfile.addItem(profile.profile())
        
        if camera := self.cp.getCurrentCamera():
            self.cmbRecordProfile.setCurrentIndex(camera.systemTabSettings.getRecordProfile())

        self.cmbRecordProfile.currentIndexChanged.connect(self.cmbRecordProfileChanged)
        self.syncGUI()
        self.setEnabled(True)

    def syncGUI(self):
        if camera := self.cp.getCurrentCamera():
            self.chkRemoteSnapshot.setChecked(camera.systemTabSettings.remote_snapshot)
            self.grpRecord.setChecked(camera.systemTabSettings.record_enable)
            if camera.systemTabSettings.record_always:
                self.radRecordAlways.setChecked(True)
                self.radRecordOnAlarm.setChecked(False)
            if camera.systemTabSettings.record_alarm:
                self.radRecordOnAlarm.setChecked(True)
                self.radRecordAlways.setChecked(False)
            self.grpSounds.setChecked(camera.systemTabSettings.sound_alarm_enable)
            if camera.systemTabSettings.sound_alarm_once:
                self.radSoundOnce.setChecked(True)
                self.radSoundLoop.setChecked(False)
            if camera.systemTabSettings.sound_alarm_loop:
                self.radSoundLoop.setChecked(True)
                self.radSoundOnce.setChecked(False)

            self.cp.btnRecord.setEnabled(not (self.grpRecord.isChecked() and self.radRecordAlways.isChecked()))
            if camera.isRecording():
                self.cp.btnRecord.setStyleSheet(self.cp.getButtonStyle("recording"))
            else:
                self.cp.btnRecord.setStyleSheet(self.cp.getButtonStyle("record"))

            if not camera.hasAudio():
                self.chkRecordAudio.setChecked(False)
                self.chkRecordAudio.setEnabled(False)
                return

            matchedProfiles = True
            if recordProfile := camera.getRecordProfile():
                self.chkRecordAudio.stateChanged.disconnect()
                self.chkRecordAudio.setChecked(not recordProfile.getDisableAudio())
                self.chkRecordAudio.stateChanged.connect(self.chkRecordAudioClicked)
                if displayProfile := camera.getDisplayProfile():
                    if recordProfile.uri() != displayProfile.uri():
                        matchedProfiles = False

            self.chkRecordAudio.setEnabled(not matchedProfiles)
            self.chkRecordAudio.setChecked(not recordProfile.getDisableAudio())

    def btnRebootClicked(self):
        if camera := self.cp.getCurrentCamera():
            result = QMessageBox.question(self, "Warning", f'{camera.name()}: Please confirm reboot')
            if result == QMessageBox.StandardButton.Yes:
                if self.cp.mw.settingsPanel.proxy.proxyType == ProxyType.CLIENT:
                    arg = "REBOOT\n\n" + camera.onvif_data.toJSON() + "\r\n"
                    self.cp.mw.client.transmit(bytearray(arg, 'utf-8'))
                else:
                    camera.onvif_data.startReboot()

    def btnSyncTimeClicked(self):
        if camera := self.cp.getCurrentCamera():
            self.dlgTimeDialog.exec(camera.onvif_data)

    def btnBrowserClicked(self):
        if camera := self.cp.lstCamera.currentItem():
            host = "http://" + camera.onvif_data.host()
            webbrowser.get().open(host)



#/********************************************************************
# Cayenue/cayenue/panels/file/treeview.py 
#
# Copyright (c) 2025  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

import os
from PyQt6.QtWidgets import QTreeView
from PyQt6.QtCore import Qt

class TreeView(QTreeView):
    def __init__(self, mw):
        super().__init__()
        self.mw = mw

    def keyPressEvent(self, event):

        pass_through = True

        match event.key():
            case Qt.Key.Key_Return:
                index = self.currentIndex()
                if index.isValid():
                    fileInfo = self.model().fileInfo(index)
                    if fileInfo.isFile():
                        if self.model().isReadOnly():
                            self.mw.filePanel.control.btnPlayClicked()
                    else:
                        if self.isExpanded(index):
                            self.collapse(index)
                        else:
                            self.expand(index)

            case Qt.Key.Key_Space:
                index = self.currentIndex()
                if index.isValid():
                    fileInfo = self.model().fileInfo(index)
                    if fileInfo.isFile():
                        if self.model().isReadOnly():
                            if player := self.mw.filePanel.getCurrentlyPlayingFile():
                                player.togglePaused()
                        self.mw.filePanel.control.setBtnPlay()

            case Qt.Key.Key_Escape:
                if self.model().isReadOnly():
                    self.mw.filePanel.control.btnStopClicked()
                else:
                    self.model().setReadOnly(True)
        
            case Qt.Key.Key_F1:
                self.mw.filePanel.onMenuInfo()

            case Qt.Key.Key_F2:
                self.mw.filePanel.onMenuRename()

            case Qt.Key.Key_Delete:
                self.mw.filePanel.onMenuRemove()

            case Qt.Key.Key_Left:
                self.mw.filePanel.rewind()
                pass_through = False

            case Qt.Key.Key_Right:
                self.mw.filePanel.fastForward()
                pass_through = False
        
        if pass_through:
            return super().keyPressEvent(event)

    def currentChanged(self, newItem, oldItem):
        if newItem.data():
            fullPath = os.path.join(self.model().rootPath(), newItem.data())
            if os.path.isfile(fullPath):
                if player := self.mw.pm.getPlayer(str(fullPath)):
                    self.mw.glWidget.focused_uri = player.uri
            self.scrollTo(self.currentIndex())



#********************************************************************
# Cayenue/cayenue/protocols/client.py
#
# Copyright (c) 2025  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

from PyQt6.QtWidgets import QMessageBox
from PyQt6.QtCore import pyqtSignal, QObject
import libonvif as onvif
from loguru import logger
from pathlib import Path

class ClientProtocolSignals(QObject):
    error = pyqtSignal(str)

class ClientProtocols():
    def __init__(self, mw):
        self.mw = mw
        self.signals = ClientProtocolSignals()
        self.signals.error.connect(self.showMsgBox)
        self.msg_box = QMessageBox(mw)
        self.msg_box.setIcon(QMessageBox.Icon.Critical)
        self.msg_box.setWindowTitle("client error") 
        self.msg_box.setText("Unable to complete request") 
        self.msg_box.setStandardButtons(QMessageBox.StandardButton.Ok) 
        self.msg_box.setMinimumWidth(650)
    def callback(self, arg):
        try:
            index = bytearray(arg).find(b'\r\n')
            msg = bytearray(arg[:index]).decode('utf-8')
            payload_size = (len(arg) - index - 2)
            configs = msg.split("\n\n")
            cmd = configs.pop(0)

            if cmd == "GET CAMERAS":
                for config in configs:
                    if len(config):
                        profiles = config.split("\n")
                        onvif_data = None
                        for idx, profile in enumerate(profiles):
                            if idx == 0:
                                onvif_data = onvif.Data(profile)
                            data = onvif.Data(profile)
                            onvif_data.addProfile(data)
                        if onvif_data:
                            self.mw.cameraPanel.getProxyData(onvif_data)
                self.mw.viewer_cameras_filled = True

            if cmd == "UPDATE":
                data = onvif.Data(configs[0])
                if camera := self.mw.cameraPanel.getCameraBySerialNumber(data.serial_number()):
                    camera.syncData(data)

            if cmd == "SNAPSHOT":
                filename = configs[0]
                if payload_size:
                    index += 2
                    with open(filename, 'wb') as file:
                        file.write(bytearray(arg[index:]))
                else:
                    path = Path(filename)
                    camera_name = path.parent.name
                    if camera := self.mw.cameraPanel.getCameraByName(camera_name):
                        if profile := camera.getDisplayProfile():
                            if player := self.mw.pm.getPlayer(profile.uri()):
                                if player.image:
                                    player.image.save(filename)

        except Exception as ex:
            print("EXCEPTION ", ex)
            return

    def error(self, msg):
        logger.error(f'Client protocol error: {msg}')
        #self.signals.error.emit(msg)

    def showMsgBox(self, msg):
        #QMessageBox.critical(self.mw, "Unable to complete request", msg)
        self.msg_box.setText(msg) 
        self.msg_box.exec()



#********************************************************************
# Cayenue/cayenue/components/progress.py
#
# Copyright (c) 2023  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

from PyQt6.QtWidgets import QSlider, QLabel, QWidget, QGridLayout
from PyQt6.QtCore import Qt, QPoint
from PyQt6.QtGui import QPainter

class Slider(QSlider):
    def __init__(self, o, P):
        super().__init__(o)
        self.P = P
        if P.showPosition:
            self.setMouseTracking(True)
        self.setFocusPolicy(Qt.FocusPolicy.NoFocus)

    def leaveEvent(self, e):
        self.P.updatePosition(-1, 0)

    def mousePressEvent(self, e):
        pct = e.position().x() / self.width()
        uri = self.P.mw.glWidget.focused_uri
        player = self.P.mw.pm.getPlayer(uri)
        if player is not None:
            player.seek(pct)

    def mouseMoveEvent(self, e):
        x = e.position().x()
        pct = x / self.width()
        self.P.updatePosition(pct, x)

class Position(QLabel):
    def __init__(self):
        super().__init__()
        self.pos = 0

    def setText(self, s, n):
        self.pos = n
        super().setText(s)

    def paintEvent(self, e):
        painter = QPainter(self)
        rect = self.fontMetrics().boundingRect(self.text())
        x = min(self.width() - rect.width(), self.pos)
        painter.drawText(QPoint(int(x), self.height()), self.text())

class Progress(QWidget):
    def __init__(self, mw):
        super().__init__()
        self.mw = mw
        self.duration = 0
        self.showPosition = True

        self.sldProgress = Slider(Qt.Orientation.Horizontal, self)
        self.sldProgress.setMaximum(1000)
        self.lblProgress = QLabel("0:00")
        self.setLabelWidth(self.lblProgress)
        self.lblDuration = QLabel("0:00")
        self.setLabelWidth(self.lblDuration)
        if self.showPosition:
            self.lblPosition = Position()

        lytProgress = QGridLayout(self)
        if self.showPosition:
            lytProgress.addWidget(self.lblPosition,  0, 1, 1, 1)
        lytProgress.addWidget(self.lblProgress,      1, 0, 1, 1)
        lytProgress.addWidget(self.sldProgress,      1, 1, 1, 1)
        lytProgress.addWidget(self.lblDuration,      1, 2, 1, 1)
        lytProgress.setContentsMargins(0, 0, 0, 0)
        lytProgress.setColumnStretch(1, 10)

    def setLabelWidth(self, l):
        l.setFixedWidth(l.fontMetrics().boundingRect("00:00:00").width())
        l.setAlignment(Qt.AlignmentFlag.AlignCenter)

    def updateDuration(self, n):
        self.duration = n
        self.lblDuration.setText(self.timestring(n))

    def updateProgress(self, f):
        self.sldProgress.setValue(int(f * 1000))
        self.lblProgress.setText(self.timestring(int(self.duration * f)))

    def updatePosition(self, f, n):
        if self.showPosition:
            if f >= 0:
                position = int(f * self.duration)
                self.lblPosition.setText(self.timestring(position), n)
            else:
                self.lblPosition.setText("", 0)

    def timestring(self, n):
        time_interval = int(n / 1000)
        hours = int(time_interval / 3600)
        minutes = int ((time_interval - (hours * 3600)) / 60)
        seconds = int ((time_interval - (hours * 3600) - (minutes * 60)))
        if hours > 0:
            buf = "%02d:%02d:%02d" % (hours, minutes, seconds)
        else:
            buf = "%d:%02d" % (minutes, seconds)
        return buf




#/********************************************************************
# Cayenue/cayenue/panels/cameras/snapshot.py 
#
# Copyright (c) 2026  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

from loguru import logger
from cayenue.enums import ProxyType, SnapshotAuth
import requests
from requests.auth import HTTPDigestAuth
from urllib.parse import urlparse, parse_qs

class Snapshot():
    def __init__(self, mw):
        self.mw = mw

    def getSnapshot(self, profile, filename, auth_type):
        parsed = urlparse(profile.snapshot_uri())

        if auth_type == SnapshotAuth.NOT_WORKING:
            return False

        if auth_type == SnapshotAuth.UNKNOWN:
            if self.getSnapshot(profile, filename, SnapshotAuth.BASIC):
                camera = self.mw.cameraPanel.getCamera(profile.uri())
                camera.setSnapshotAuth(SnapshotAuth.BASIC)
                logger.debug(f"camera {camera.name()} snapshot auth type set to BASIC")
                return True

            else:
                if self.getSnapshot(profile, filename, SnapshotAuth.DIGEST):
                    camera = self.mw.cameraPanel.getCamera(profile.uri())
                    camera.setSnapshotAuth(SnapshotAuth.DIGEST)
                    logger.debug(f"camera {camera.name()} snapshot auth type set to DIGEST")
                    return True

        if auth_type == SnapshotAuth.BASIC:
            try:
                simple_url = f"{parsed.scheme}://{profile.username()}:{profile.password()}@{parsed.netloc}{parsed.path}?{parsed.query}"
                response = requests.get(simple_url, timeout=5)
                if response.status_code == 200:
                    with open(filename, 'wb') as file:
                        file.write(response.content)
                    #logger.debug(f"Image downloaded successfully as {filename}")
                    return True
                else:
                    logger.debug(f"get snapshot BASIC auth status code: {response}")
                    return False
            except Exception as ex:
                logger.debug(f'set snapshot BASIC auth exception {ex}')
            return False

        if auth_type == SnapshotAuth.DIGEST:
            try:
                base_url = f"{parsed.scheme}://{parsed.netloc}{parsed.path}"
                params = {k: v[0] for k, v in parse_qs(parsed.query).items()}
                response = requests.get(base_url, params=params, auth=HTTPDigestAuth(profile.username(), profile.password()), timeout=5)
                if response.status_code == 200:
                    with open(filename, 'wb') as file:
                        file.write(response.content)
                    #logger.debug(f"Image downloaded successfully as {filename}")
                    return True
                else:
                    logger.debug(f"get snapshot DIGEST auth status code: {response}")
                    return False
            except Exception as ex:
                logger.debug(f'set snapshot DIGEST auth exception {ex}')
            return False

        return False

    def getSnapshotBuffered(self, profile, buffer, auth_type):
        print("Here we are at getSnapshotBuffered", auth_type)
        parsed = urlparse(profile.snapshot_uri())

        if auth_type == SnapshotAuth.NOT_WORKING:
            return False

        if auth_type == SnapshotAuth.UNKNOWN:
            if self.getSnapshotBuffered(profile, buffer, SnapshotAuth.BASIC):
                #camera = self.mw.cameraPanel.getCamera(profile.uri())
                #camera.setSnapshotAuth(SnapshotAuth.BASIC)
                #logger.debug(f"camera {camera.name()} snapshot auth type set to BASIC")
                return True

            else:
                if self.getSnapshotBuffered(profile, buffer, SnapshotAuth.DIGEST):
                    #camera = self.mw.cameraPanel.getCamera(profile.uri())
                    #camera.setSnapshotAuth(SnapshotAuth.DIGEST)
                    #logger.debug(f"camera {camera.name()} snapshot auth type set to DIGEST")
                    return True

        if auth_type == SnapshotAuth.BASIC:
            try:
                simple_url = f"{parsed.scheme}://{profile.username()}:{profile.password()}@{parsed.netloc}{parsed.path}?{parsed.query}"
                response = requests.get(simple_url, timeout=5)
                if response.status_code == 200:
                    
                    print("Writing buffer from BASIC AUTH")
                    buffer.write(response.content)
                    
                    #logger.debug(f"Image downloaded successfully as {filename}")
                    return True
                else:
                    logger.debug(f"get snapshot BASIC auth status code: {response}")
                    return False
            except Exception as ex:
                logger.debug(f'set snapshot BASIC auth exception {ex}')
            return False

        if auth_type == SnapshotAuth.DIGEST:
            try:
                base_url = f"{parsed.scheme}://{parsed.netloc}{parsed.path}"
                print("BASE URL:", base_url)
                params = {k: v[0] for k, v in parse_qs(parsed.query).items()}
                response = requests.get(base_url, params=params, auth=HTTPDigestAuth(profile.username(), profile.password()), timeout=5)
                if response.status_code == 200:
                    
                    print("Writing buffer from DIGEST AUTH")
                    buffer.write(response.content)
                    #with open("/home/stephen/Pictures/sample.jpg", "wb") as file:
                        #file.write(response.content)
                        #file.write(buffer.getvalue())
                    #print("where is the file")

                    #logger.debug(f"Image downloaded successfully as {filename}")
                    return True
                else:
                    logger.debug(f"get snapshot DIGEST auth status code: {response}")
                    return False
            except Exception as ex:
                logger.debug(f'set snapshot DIGEST auth exception {ex}')
            return False

        return False

    def getBufferedSnapshot(self, profile, buffer, camera):
        camera.setSnapshotAuth(SnapshotAuth.UNKNOWN)
        print("call came into getBufferedSnapshot", camera.name(), camera.snapshotAuth)
        if not self.getSnapshotBuffered(profile, buffer, camera.snapshotAuth):
            if camera.snapshotAuth is not SnapshotAuth.NOT_WORKING:
                logger.debug(f'camera {camera.name()} snapshot auth type set to NOT_WORKING')
                camera.setSnapshotAuth(SnapshotAuth.NOT_WORKING)
            #player.image.save(filename)
            #return buffer


    def __call__(self, profile, filename, camera, player):
        if not camera.systemTabSettings.remote_snapshot:
            player.image.save(filename)
            return
        if self.mw.settingsPanel.proxy.proxyType == ProxyType.CLIENT:
            print("CLIENT PROXY", camera.name())
            profile.setUserData(filename)
            self.mw.client.transmit(bytearray(f"SNAPSHOT\n\n{profile.toJSON()}\r\n", 'utf-8'))
            return
        if not self.getSnapshot(profile, filename, camera.snapshotAuth):
            if camera.snapshotAuth is not SnapshotAuth.NOT_WORKING:
                logger.debug(f'camera {camera.name()} snapshot auth type set to NOT_WORKING')
                camera.setSnapshotAuth(SnapshotAuth.NOT_WORKING)
            player.image.save(filename)
            #logger.debug(f'Raw stream image saved as {filename}')



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Copyright (c) Megvii Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#******************************************************************************/

import torch
import torch.nn as nn

from .darknet import CSPDarknet
from .network_blocks import BaseConv, CSPLayer, DWConv


class YOLOPAFPN(nn.Module):
    """
    YOLOv3 model. Darknet 53 is the default backbone of this model.
    """

    def __init__(
        self,
        depth=1.0,
        width=1.0,
        in_features=("dark3", "dark4", "dark5"),
        in_channels=[256, 512, 1024],
        depthwise=False,
        act="silu",
    ):
        super().__init__()
        self.backbone = CSPDarknet(depth, width, depthwise=depthwise, act=act)
        self.in_features = in_features
        self.in_channels = in_channels
        Conv = DWConv if depthwise else BaseConv

        self.upsample = nn.Upsample(scale_factor=2, mode="nearest")
        self.lateral_conv0 = BaseConv(
            int(in_channels[2] * width), int(in_channels[1] * width), 1, 1, act=act
        )
        self.C3_p4 = CSPLayer(
            int(2 * in_channels[1] * width),
            int(in_channels[1] * width),
            round(3 * depth),
            False,
            depthwise=depthwise,
            act=act,
        )  # cat

        self.reduce_conv1 = BaseConv(
            int(in_channels[1] * width), int(in_channels[0] * width), 1, 1, act=act
        )
        self.C3_p3 = CSPLayer(
            int(2 * in_channels[0] * width),
            int(in_channels[0] * width),
            round(3 * depth),
            False,
            depthwise=depthwise,
            act=act,
        )

        # bottom-up conv
        self.bu_conv2 = Conv(
            int(in_channels[0] * width), int(in_channels[0] * width), 3, 2, act=act
        )
        self.C3_n3 = CSPLayer(
            int(2 * in_channels[0] * width),
            int(in_channels[1] * width),
            round(3 * depth),
            False,
            depthwise=depthwise,
            act=act,
        )

        # bottom-up conv
        self.bu_conv1 = Conv(
            int(in_channels[1] * width), int(in_channels[1] * width), 3, 2, act=act
        )
        self.C3_n4 = CSPLayer(
            int(2 * in_channels[1] * width),
            int(in_channels[2] * width),
            round(3 * depth),
            False,
            depthwise=depthwise,
            act=act,
        )

    def forward(self, input):
        """
        Args:
            inputs: input images.

        Returns:
            Tuple[Tensor]: FPN feature.
        """

        #  backbone
        out_features = self.backbone(input)
        features = [out_features[f] for f in self.in_features]
        [x2, x1, x0] = features

        fpn_out0 = self.lateral_conv0(x0)  # 1024->512/32
        f_out0 = self.upsample(fpn_out0)  # 512/16
        f_out0 = torch.cat([f_out0, x1], 1)  # 512->1024/16
        f_out0 = self.C3_p4(f_out0)  # 1024->512/16

        fpn_out1 = self.reduce_conv1(f_out0)  # 512->256/16
        f_out1 = self.upsample(fpn_out1)  # 256/8
        f_out1 = torch.cat([f_out1, x2], 1)  # 256->512/8
        pan_out2 = self.C3_p3(f_out1)  # 512->256/8

        p_out1 = self.bu_conv2(pan_out2)  # 256->256/16
        p_out1 = torch.cat([p_out1, fpn_out1], 1)  # 256->512/16
        pan_out1 = self.C3_n3(p_out1)  # 512->512/16

        p_out0 = self.bu_conv1(pan_out1)  # 512->512/32
        p_out0 = torch.cat([p_out0, fpn_out0], 1)  # 512->1024/32
        pan_out0 = self.C3_n4(p_out0)  # 1024->1024/32

        outputs = (pan_out2, pan_out1, pan_out0)
        return outputs



#********************************************************************
# Cayenue/cayenue/protocols/listen.py
#
# Copyright (c) 2025  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

from loguru import logger
from time import sleep
from datetime import datetime

class Detection():
    def __init__(self, boxes, alarm, width, height, timestamp):
        self.boxes = boxes
        self.alarm = alarm
        self.width = width
        self.height = height
        self.timestamp = timestamp

class ListenProtocols():
    def __init__(self, mw):
        self.mw = mw
        self.cameras = {}
        self.thread_lock = False
        self.detections = {}
        self.last_timestamp = ""

    def error(self, msg):
        if msg.find("WSACancelBlockingCall") < 0:
            logger.error(msg)

    def lock(self):
        while self.thread_lock:
            sleep(0.001)
        self.thread_lock = True

    def unlock(self):
        self.thread_lock = False

    def getDetection(self, uri):
        result = None
        self.lock()
        result = self.detections.get(uri)
        self.unlock()
        return result
    
    def setDetection(self, uri, detection):
        self.lock()
        self.detections[uri] = detection
        self.unlock()

    def callback(self, msg):
        arguments = msg.split("\n\n")
        timestamp = arguments.pop(0)
        if timestamp == self.last_timestamp:
            return
        
        self.last_timestamp = timestamp
        cmd = arguments.pop(0)

        if cmd == "ALARMS":
            self.mw.alarm_states = arguments
            self.mw.last_alarm = datetime.now()



from .camera.camerapanel import CameraPanel
from .file.filepanel import FilePanel
from .video.videopanel import VideoPanel
from .settings.settingspanel import SettingsPanel
from .audio.audiopanel import AudioPanel
from .picture.picturepanel import PicturePanel



#*******************************************************************************
# Cayenue/cayenue/components/directoryselector.py
#
# Copyright (c) 2023 Stephen Rhodes 
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#******************************************************************************/

from PyQt6.QtWidgets import QWidget, QLineEdit, QPushButton, \
    QLabel, QGridLayout, QFileDialog
from PyQt6.QtCore import pyqtSignal, QObject
import sys

class DirectorySelectorSignals(QObject):
    dirChanged = pyqtSignal(str)

class DirectorySelector(QWidget):
    def __init__(self, mw, name, label, location=""):
        super().__init__()
        self.mw = mw
        self.directoryKey = name + "/directory"
        self.signals = DirectorySelectorSignals()

        self.txtDirectory = QLineEdit()
        self.txtDirectory.setText(self.mw.settings.value(self.directoryKey, location))
        self.txtDirectory.setEnabled(False)
        self.btnSelect = QPushButton("...")
        self.btnSelect.clicked.connect(self.btnSelectClicked)
        lblSelect = QLabel(label)

        lytMain = QGridLayout(self)
        lytMain.setContentsMargins(0, 0, 0, 0)
        if len(label):
            lytMain.addWidget(lblSelect,       0, 0, 1, 1)
        lytMain.addWidget(self.txtDirectory,   0, 1, 1, 1)
        lytMain.addWidget(self.btnSelect,      0, 2, 1, 1)
        lytMain.setColumnStretch(1, 10)
        self.setContentsMargins(0, 0, 0, 0)

    def btnSelectClicked(self):
        path = None
        if sys.platform == "linux":
            path = QFileDialog.getExistingDirectory(self, "Select Directory", self.txtDirectory.text(), QFileDialog.Option.DontUseNativeDialog)
        else:
            path = QFileDialog.getExistingDirectory(self, "Select Directory", self.txtDirectory.text())
        if path:
            self.txtDirectory.setText(path)
            self.txtDirectoryChanged(path)

    def txtDirectoryChanged(self, text):
        self.mw.settings.setValue(self.directoryKey, text)
        self.signals.dirChanged.emit(text)

    def text(self):
        return self.txtDirectory.text()
    
    def setEnabled(self, arg):
        self.txtDirectory.setEnabled(arg)
        self.btnSelect.setEnabled(arg)
        super().setEnabled(arg)




#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#******************************************************************************/

import torch
from torch import nn
from torch.hub import load_state_dict_from_url

__all__ = [
    "create_yolox_model",
    "yolox_nano",
    "yolox_tiny",
    "yolox_s",
    "yolox_m",
    "yolox_l",
    "yolox_x",
    "yolov3",
    "yolox_custom"
]

_CKPT_ROOT_URL = "https://github.com/Megvii-BaseDetection/YOLOX/releases/download"
_CKPT_FULL_PATH = {
    "yolox-nano": f"{_CKPT_ROOT_URL}/0.1.1rc0/yolox_nano.pth",
    "yolox-tiny": f"{_CKPT_ROOT_URL}/0.1.1rc0/yolox_tiny.pth",
    "yolox-s": f"{_CKPT_ROOT_URL}/0.1.1rc0/yolox_s.pth",
    "yolox-m": f"{_CKPT_ROOT_URL}/0.1.1rc0/yolox_m.pth",
    "yolox-l": f"{_CKPT_ROOT_URL}/0.1.1rc0/yolox_l.pth",
    "yolox-x": f"{_CKPT_ROOT_URL}/0.1.1rc0/yolox_x.pth",
    "yolov3": f"{_CKPT_ROOT_URL}/0.1.1rc0/yolox_darknet.pth",
}


def create_yolox_model(name: str, pretrained: bool = True, num_classes: int = 80, device=None,
                       exp_path: str = None, ckpt_path: str = None) -> nn.Module:
    """creates and loads a YOLOX model

    Args:
        name (str): name of model. for example, "yolox-s", "yolox-tiny" or "yolox_custom"
        if you want to load your own model.
        pretrained (bool): load pretrained weights into the model. Default to True.
        device (str): default device to for model. Default to None.
        num_classes (int): number of model classes. Default to 80.
        exp_path (str): path to your own experiment file. Required if name="yolox_custom"
        ckpt_path (str): path to your own ckpt. Required if name="yolox_custom" and you want to
            load a pretrained model


    Returns:
        YOLOX model (nn.Module)
    """
    from yolox.exp import get_exp, Exp

    if device is None:
        device = "cuda:0" if torch.cuda.is_available() else "cpu"
    device = torch.device(device)

    assert name in _CKPT_FULL_PATH or name == "yolox_custom", \
        f"user should use one of value in {_CKPT_FULL_PATH.keys()} or \"yolox_custom\""
    if name in _CKPT_FULL_PATH:
        exp: Exp = get_exp(exp_name=name)
        exp.num_classes = num_classes
        yolox_model = exp.get_model()
        if pretrained and num_classes == 80:
            weights_url = _CKPT_FULL_PATH[name]
            ckpt = load_state_dict_from_url(weights_url, map_location="cpu")
            if "model" in ckpt:
                ckpt = ckpt["model"]
            yolox_model.load_state_dict(ckpt)
    else:
        assert exp_path is not None, "for a \"yolox_custom\" model exp_path must be provided"
        exp: Exp = get_exp(exp_file=exp_path)
        yolox_model = exp.get_model()
        if ckpt_path:
            ckpt = torch.load(ckpt_path, map_location="cpu")
            if "model" in ckpt:
                ckpt = ckpt["model"]
            yolox_model.load_state_dict(ckpt)

    yolox_model.to(device)
    return yolox_model


def yolox_nano(pretrained: bool = True, num_classes: int = 80, device: str = None) -> nn.Module:
    return create_yolox_model("yolox-nano", pretrained, num_classes, device)


def yolox_tiny(pretrained: bool = True, num_classes: int = 80, device: str = None) -> nn.Module:
    return create_yolox_model("yolox-tiny", pretrained, num_classes, device)


def yolox_s(pretrained: bool = True, num_classes: int = 80, device: str = None) -> nn.Module:
    return create_yolox_model("yolox-s", pretrained, num_classes, device)


def yolox_m(pretrained: bool = True, num_classes: int = 80, device: str = None) -> nn.Module:
    return create_yolox_model("yolox-m", pretrained, num_classes, device)


def yolox_l(pretrained: bool = True, num_classes: int = 80, device: str = None) -> nn.Module:
    return create_yolox_model("yolox-l", pretrained, num_classes, device)


def yolox_x(pretrained: bool = True, num_classes: int = 80, device: str = None) -> nn.Module:
    return create_yolox_model("yolox-x", pretrained, num_classes, device)


def yolov3(pretrained: bool = True, num_classes: int = 80, device: str = None) -> nn.Module:
    return create_yolox_model("yolov3", pretrained, num_classes, device)


def yolox_custom(ckpt_path: str = None, exp_path: str = None, device: str = None) -> nn.Module:
    return create_yolox_model("yolox_custom", ckpt_path=ckpt_path, exp_path=exp_path, device=device)



#********************************************************************
# Cayenue/cayenue/player.py
#
# Copyright (c) 2025  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

from datetime import datetime
from pathlib import Path
from PyQt6.QtGui import QPainter, QColorConstants
from PyQt6.QtCore import pyqtSignal, QObject, QTimer, QRectF
from collections import deque
import avio
import time
import os
import pathlib
from loguru import logger
from datetime import datetime
import threading
from cayenue.panels.camera import Snapshot

class PlayerSignals(QObject):
    start = pyqtSignal()
    stop = pyqtSignal()
    play_alarm_sound = pyqtSignal(str)

class Player(avio.Player):
    def __init__(self, uri, mw):
        super().__init__(uri)
        self.mw = mw
        self.signals = PlayerSignals()
        self.ary = None
        self.image = None
        self.desired_aspect = 0
        self.analyze_video = False
        self.analyze_audio = False
        self.videoModelSettings = None
        self.audioModelSettings = None
        self.detection_count = deque()
        self.last_image = None
        self.last_render = None
        self.needs_render = False
        self.timer = None
        self.thread_lock = False
        self.marked_for_death = False

        self.boxes = None
        self.labels = []
        self.scores = []
        self.last_alarm_sound = datetime.now()

        self.save_image_filename = None
        self.output_file_start_time = None
        self.estimated_file_size = 0
        self.packet_drop_frame_counter = 0
        self.last_msg = ""

        self.alarm_state = 0
        self.last_alarm_state = 0
        self.file_progress = 0.0

        self.snapshot = Snapshot(mw)

        if (len(uri)):
            self.timer = QTimer()
            self.timer.setInterval(self.mw.settingsPanel.alarm.spnLagTime.value() * 1000)
            self.timer.setSingleShot(True)
            self.timer.timeout.connect(self.timeout)
            self.signals.start.connect(self.timer.start)
            self.signals.stop.connect(self.timer.stop)
            self.signals.play_alarm_sound.connect(self.soundAlarm)

    def lock(self):
        # the lock protects the image
        if self.thread_lock:
            time.sleep(0.001)
        self.thread_lock = True

    def unlock(self):
        self.thread_lock = False

    def requestShutdown(self, reconnect=False):
        if self.isCameraStream():
            self.setAlarmState(0)
        else:
            if self.isPaused():
                self.togglePaused()
        self.analyze_video = False
        self.analyze_audio = False
        self.request_reconnect = reconnect
        self.marked_for_death = True
        self.terminate()

    def systemTabSettings(self):
        result = None
        if camera := self.mw.cameraPanel.getCamera(self.uri):
            result = camera.systemTabSettings
        return result

    def setAlarmState(self, state):
         
        self.alarm_state = int(state)

        record_enable = self.systemTabSettings().record_enable if self.systemTabSettings() else False
        record_alarm = self.systemTabSettings().record_alarm if self.systemTabSettings() else False
        if camera := self.mw.cameraPanel.getCamera(self.uri):
            manual_recording = camera.manual_recording if camera else False
            profile = camera.getRecordProfile() if camera else None
            player = self.mw.pm.getPlayer(profile.uri()) if profile else None

            if state:
                self.signals.start.emit()
                if record_enable and record_alarm:
                    if player:
                        if not player.isRecording():
                            if filename := player.getOutputFilename():
                                player.toggleRecording(filename)
                                if current_camera := self.mw.cameraPanel.getCurrentCamera():
                                    if camera.serial_number() == current_camera.serial_number():
                                        self.mw.cameraPanel.syncGUI()
            else:
                self.signals.stop.emit()
                if record_alarm and not manual_recording:
                    if player:
                        if player.isRecording():
                            player.toggleRecording("")
                            self.mw.settingsPanel.storage.signals.updateDiskUsage.emit()
                            if current_camera := self.mw.cameraPanel.getCurrentCamera():
                                if camera.serial_number() == current_camera.serial_number():
                                    self.mw.cameraPanel.syncGUI()

    def timeout(self):
        self.setAlarmState(0)

    def getOutputFilename(self):
        filename = None
        if camera := self.mw.cameraPanel.getCamera(self.uri):
            d = self.mw.settingsPanel.storage.dirArchive.txtDirectory.text()
            root = os.path.join(d, camera.text())
            Path(root).mkdir(parents=True, exist_ok=True)
            self.output_file_start_time = datetime.now()
            filename = '{0:%Y%m%d%H%M%S}'.format(self.output_file_start_time)
            filename = os.path.join(root, filename)
            self.setMetaData("title", camera.text())
        return filename

    def handleAlarm(self, state):
        if self.analyze_video or self.analyze_audio:
            if state:
                self.setAlarmState(1)

                if self.systemTabSettings():
                    save_picture = self.alarm_state != self.last_alarm_state \
                                and self.mw.settingsPanel.alarm.chkSavePicture.isChecked() \
                                and self.systemTabSettings().record_enable

                    if save_picture and self.image:
                        try:
                            saveLocal= not self.systemTabSettings().remote_snapshot
                            if saveLocal:
                                self.lock()
                                img = self.image.copy()
                                self.unlock()
                                root = os.path.join(self.mw.settingsPanel.storage.dirPictures.txtDirectory.text(), self.mw.cameraPanel.getCamera(self.uri).text())
                                pathlib.Path(root).mkdir(parents=True, exist_ok=True)
                                filename = '{0:%Y%m%d%H%M%S.jpg}'.format(datetime.now())
                                filename = os.path.join(root, filename)

                                painter_img = QPainter(img)
                                painter_img.setPen(QColorConstants.Red)
                                if self.boxes:
                                    for box in self.boxes:
                                        p = (box[0])
                                        q = (box[1])
                                        r = (box[2] - box[0])
                                        s = (box[3] - box[1])
                                        painter_img.drawRect(QRectF(p, q, r, s))
                                painter_img.end()
                                img.save(filename)
                            else:
                                root = self.mw.settingsPanel.storage.dirPictures.txtDirectory.text() + "/" + self.mw.cameraPanel.getCamera(self.uri).text()
                                Path(root).mkdir(parents=True, exist_ok=True)
                                filename = '{0:%Y%m%d%H%M%S.jpg}'.format(datetime.now())
                                filename = str(root + "/" + filename)

                                if camera := self.mw.cameraPanel.getCamera(self.uri):
                                    profile = camera.getRecordProfile()
                                    if not profile:
                                        profile = camera.getProfile(self.uri)

                                    thread = threading.Thread(target=self.snapshot, args=(profile, filename, camera, self))
                                    thread.start()

                        except Exception as ex:
                            logger.error(f'player handle alarm write file exception: {ex}')

                    if self.systemTabSettings().sound_alarm_enable:
                        filename = os.path.join(self.mw.getLocation(), "cayenue", "resources", self.mw.settingsPanel.alarm.cmbSoundFiles.currentText())
                        if self.systemTabSettings().sound_alarm_once:
                            if self.alarm_state != self.last_alarm_state:
                                self.signals.play_alarm_sound.emit(filename)
                        if self.systemTabSettings().sound_alarm_loop:
                            p = self.mw.pm.getPlayer(filename)
                            if not p:
                                self.signals.play_alarm_sound.emit(filename)

            self.last_alarm_state = self.alarm_state

        else:
            self.setAlarmState(0)

    def soundAlarm(self, filename):
        player = Player(filename, self.mw)
        player.live_stream = False
        player.mediaPlayingStopped = self.mw.pm.removePlayer
        player.audio_driver_index = self.mw.settingsPanel.general.cmbAudioDriver.currentIndex()
        player.request_reconnect = False
        player.disable_video = True
        player.setVolume(self.mw.settingsPanel.alarm.sldAlarmVolume.value())
        self.mw.pm.startPlayer(player)

    def processModelOutput(self):
        sum = 0
        if self.boxes is not None:
            if camera := self.mw.cameraPanel.getCamera(self.uri):
                if len(self.detection_count) > camera.videoModelSettings.sampleSize - 1 and len(self.detection_count):
                    self.detection_count.popleft()
                if len(self.boxes):
                    self.detection_count.append(1)
                else:
                    self.detection_count.append(0)

                for count in self.detection_count:
                    sum += count
        return sum

    def loadRemoteDetections(self):
        for idx, alarm in enumerate(self.mw.alarm_states):
            serial_number = self.mw.alarm_ordinals.get(idx, None)
            if camera := self.mw.cameraPanel.getCamera(self.uri):
                if camera.serial_number() == serial_number:
                    self.handleAlarm(int(alarm))



#*******************************************************************************
# Cayenue/cayenue/components/comboselector.py
#
# Copyright (c) 2023 Stephen Rhodes 
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#******************************************************************************/

from PyQt6.QtWidgets import QWidget, QLabel, QComboBox, QGridLayout

class ComboSelector(QWidget):
    def __init__(self, mw, name, list, select, id=None):
        super().__init__()
        self.mw = mw
        self.selectKey = "Module/"
        if id is not None:
            self.selectKey += id + "/"
        self.selectKey += name + "/filename"

        self.cmbBox = QComboBox()
        self.cmbBox.addItems(list)
        self.cmbBox.setCurrentText(self.mw.settings.value(self.selectKey, select))
        self.cmbBox.currentTextChanged.connect(self.cmbBoxChanged)

        lblBox = QLabel(name)
        lytBox = QGridLayout(self)
        lytBox.addWidget(lblBox,       0, 0, 1, 1)
        lytBox.addWidget(self.cmbBox,  0, 1, 1, 1)
        lytBox.setColumnStretch(1, 10)
        lytBox.setContentsMargins(0, 0, 0, 0)

    def cmbBoxChanged(self, text):
        self.mw.settings.setValue(self.selectKey, text)

    def currentText(self):
        return self.cmbBox.currentText()

    def clear(self):
        self.cmbBox.clear()

    def addItems(self, items):
        self.cmbBox.addItems(items)


#/********************************************************************
# Cayenue/panels/video/modules/yolox.py 
#
# Copyright (c) 2023  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

IMPORT_ERROR = ""
MODULE_NAME = "yolox"

try:
    import os
    import sys
    from loguru import logger
    import numpy as np
    from pathlib import Path
    from cayenue.components import ComboSelector, FileSelector, ThresholdSlider, TargetSelector
    from PyQt6.QtWidgets import QWidget, QGridLayout, QLabel, QCheckBox, QMessageBox, \
        QGroupBox, QSpinBox
    from cayenue.panels.video.modules.settings import VideoModelSettings
    import time
    import torch
    import requests
    from torchvision.transforms import functional
    import torch.nn as nn
    from cayenue.panels.video.modules.yolox.models import YOLOX, YOLOPAFPN, YOLOXHead
    import cayenue.panels.video.modules.yolox.utils
    if sys.platform != "darwin":
        import openvino as ov

except ModuleNotFoundError as ex:
    IMPORT_ERROR = str(ex)
    logger.debug("Import Error has occurred, missing modules need to be installed, please consult documentation: ", ex)

os.environ['KMP_DUPLICATE_LIB_OK']='True'

class VideoConfigure(QWidget):
    def __init__(self, mw):
        try:
            super().__init__()
            self.mw = mw
            self.name = MODULE_NAME
            self.camera = None
            self.initialized = False
            self.autoKey = "Module/" + MODULE_NAME + "/autoDownload"

            if len(IMPORT_ERROR):
                self.mw.videoPanel.lblCamera.setText(f'Configuration error - {IMPORT_ERROR}')
                return

            self.chkAuto = QCheckBox("Automatically download model")
            self.chkAuto.setChecked(int(self.mw.settings.value(self.autoKey, 1)))
            self.chkAuto.stateChanged.connect(self.chkAutoClicked)
            self.txtFilename = FileSelector(mw, MODULE_NAME)
            self.txtFilename.setEnabled(not self.chkAuto.isChecked())
            self.cmbRes = ComboSelector(mw, "Size", ("160", "320", "480", "640", "960", "1280"), "640", MODULE_NAME)
            self.cmbModelName = ComboSelector(mw, "Name", ("yolox_tiny", "yolox_s", "yolox_m", "yolox_l", "yolox_x"), "yolox_s", MODULE_NAME)
            
            apis = None
            if sys.platform == "darwin":
                apis = ["PyTorch"]
                default_api = "PyTorch"
            elif sys.platform == "win32":
                apis = ["OpenVINO", "PyTorch"]
                default_api = "OpenVINO"
            else:
                apis = ["PyTorch", "OpenVINO"]            
                if self.mw.haveNvidia():
                    default_api = "PyTorch"
                else:
                    default_api = "OpenVINO"
            
            self.cmbAPI = ComboSelector(mw, "API", apis, default_api, MODULE_NAME)
            self.cmbAPI.cmbBox.currentTextChanged.connect(self.cmbAPIChanged)

            self.cmbDevice = ComboSelector(mw, "Device", self.getDevices(self.cmbAPI.currentText()), "AUTO", MODULE_NAME)

            self.sldConfThre = ThresholdSlider(mw, "Confidence", MODULE_NAME)
            self.selTargets = TargetSelector(self.mw, MODULE_NAME)

            self.spnSkipFrames = QSpinBox()
            self.lblSkipFrames = QLabel("Skip Frames")
            self.spnSkipFrames.setValue(0)
            self.spnSkipFrames.valueChanged.connect(self.spnSkipFramesChanged)

            self.spnSampleSize = QSpinBox()
            self.lblSampleSize = QLabel("Sample Size")
            self.spnSampleSize.setMinimum(1)
            self.spnSampleSize.setValue(1)
            self.spnSampleSize.valueChanged.connect(self.spnSampleSizeChanged)

            grpSystem = QGroupBox("System wide model parameters")
            lytSystem = QGridLayout(grpSystem)
            lytSystem.addWidget(self.chkAuto,      0, 0, 1, 4)
            lytSystem.addWidget(self.txtFilename,  1, 0, 1, 4)
            lytSystem.addWidget(self.cmbModelName, 2, 0, 1, 2)
            lytSystem.addWidget(self.cmbRes,       2, 2, 1, 2)
            lytSystem.addWidget(self.cmbAPI,       3, 0, 1, 2)
            lytSystem.addWidget(self.cmbDevice,    3, 2, 1, 2)

            self.grpCamera = QGroupBox("Check to enable")
            self.grpCamera.setCheckable(True)
            self.grpCamera.clicked.connect(self.mw.cameraPanel.tabVideo.chkAnalyzeVideoChecked)
            lytCamera = QGridLayout(self.grpCamera)
            lytCamera.addWidget(self.sldConfThre,    0, 0, 1, 4)
            lytCamera.addWidget(self.lblSkipFrames,  1, 0, 1, 1)
            lytCamera.addWidget(self.spnSkipFrames,  1, 1, 1, 1)
            lytCamera.addWidget(self.lblSampleSize,  1, 2, 1, 1)
            lytCamera.addWidget(self.spnSampleSize,  1, 3, 1, 1)
            lytCamera.addWidget(QLabel(),            2, 0, 1, 4)
            lytCamera.addWidget(self.selTargets,     3, 0, 1, 4)

            lytMain = QGridLayout(self)
            lytMain.addWidget(grpSystem,         0, 0, 1, 1)
            lytMain.addWidget(QLabel(),          1, 0, 1, 1)
            lytMain.addWidget(self.grpCamera,    2, 0, 1, 1)
            lytMain.addWidget(QLabel(),          3, 0, 1, 1)
            lytMain.setRowStretch(3, 10)

            self.enableControls(False)
            if camera := self.mw.cameraPanel.getCurrentCamera():
                self.setCamera(camera)

            self.initialized = True

        except Exception as ex:
            logger.exception(MODULE_NAME + " configure failed to load")
            QMessageBox.critical(None, f'{MODULE_NAME} Error', f'{MODULE_NAME} configure failed to initialize: {ex}')

    def chkAutoClicked(self, state):
        self.mw.settings.setValue(self.autoKey, state)
        self.txtFilename.setEnabled(not self.chkAuto.isChecked())

    def spnSkipFramesChanged(self, value):
        if self.camera:
            if self.camera.videoModelSettings:
                self.camera.videoModelSettings.setSkipFrames(value)

    def spnSampleSizeChanged(self, value):
        if self.camera:
            if self.camera.videoModelSettings:
                self.camera.videoModelSettings.setSampleSize(value)

        self.selTargets.sldGain.setMaximum(value)

    def getDevices(self, api):
        devices = []
        if api == "OpenVINO" and sys.platform != "darwin":
            devices = ["AUTO"] + ov.Core().available_devices
        if api == "PyTorch":
            devices = ["auto", "cpu"]
            if torch.cuda.is_available():
                devices.append("cuda")
            if torch.backends.mps.is_available():
                devices.append("mps")
        return devices

    def cmbAPIChanged(self, text):
        self.cmbDevice.clear()
        self.cmbDevice.addItems(self.getDevices(text))

    def setCamera(self, camera):
        if not camera:
            return
        
        self.camera = camera
        if camera.videoModelSettings.module_name != MODULE_NAME:
            camera.videoModelSettings = VideoModelSettings(self.mw, camera, MODULE_NAME)

        if not len(IMPORT_ERROR):
            self.mw.videoPanel.lblCamera.setText(f'Camera - {camera.name()}')
            self.sldConfThre.setValue(camera.videoModelSettings.confidence)
            self.spnSkipFrames.setValue(camera.videoModelSettings.skipFrames)
            self.spnSampleSize.setValue(camera.videoModelSettings.sampleSize)
            self.selTargets.setModelParameters(camera.videoModelSettings)
            if profile := self.mw.cameraPanel.getProfile(camera.uri()):
                self.enableControls(profile.getAnalyzeVideo())

    def enableControls(self, state):
        self.grpCamera.setChecked(bool(state))
        self.grpCamera.setEnabled(False)

        if camera := self.mw.cameraPanel.getCurrentCamera():
            self.grpCamera.setEnabled(True)
            if state:
                self.grpCamera.setTitle("Video analysis enabled")
            else:
                self.grpCamera.setTitle("Check to enable")

class VideoWorker:
    def __init__(self, mw):
        try:
            self.mw = mw
            self.last_ex = ""
            self.api = self.mw.videoConfigure.cmbAPI.currentText()

            if self.mw.videoConfigure.name != MODULE_NAME or len(IMPORT_ERROR) > 0 or self.mw.glWidget.model_loading:
                return
            
            self.mw.glWidget.model_loading = True
            time.sleep(1)

            self.mw.videoConfigure.selTargets.indAlarm.setState(0)
            self.mw.videoConfigure.selTargets.barLevel.setLevel(0)

            self.torch_device = None
            self.torch_device_name = None
            self.ov_device = None
            self.compiled_model = None
            ov_model = None

            self.num_classes = 80
            self.res = int(self.mw.videoConfigure.cmbRes.currentText())
            initializer_data = torch.rand(1, 3, self.res, self.res)
            self.model_name = self.mw.videoConfigure.cmbModelName.currentText()

            if self.api == "PyTorch":
                self.torch_device_name = self.mw.videoConfigure.cmbDevice.currentText()
            if self.api == "OpenVINO":
                self.ov_device = self.mw.videoConfigure.cmbDevice.currentText()

            if self.api == "OpenVINO" and Path(self.get_ov_model_filename()).is_file() and sys.platform != "darwin":
                ov_model = ov.Core().read_model(self.get_ov_model_filename())
            
            if (self.api == "OpenVINO" and not ov_model) or self.api == "PyTorch":
                self.torch_device_name = "cpu"
                if self.api == "PyTorch":
                    if torch.cuda.is_available():
                        self.torch_device_name = "cuda"
                    if torch.backends.mps.is_available():
                        self.torch_device_name = "mps"
                    if self.mw.videoConfigure.cmbDevice.currentText() == "cpu":
                        self.torch_device_name = "cpu"
                self.torch_device = torch.device(self.torch_device_name)

                size = {'yolox_tiny': [0.33, 0.375],
                        'yolox_s': [0.33, 0.50], 
                        'yolox_m': [0.67, 0.75],
                        'yolox_l': [1.00, 1.00],
                        'yolox_x': [1.33, 1.25]}[self.model_name]

                self.model = None
                self.model = self.get_model(self.num_classes, size[0], size[1], None).to(self.torch_device)
                self.model.eval()

                self.ckpt_file = None
                if self.mw.videoConfigure.chkAuto.isChecked():
                    self.ckpt_file = self.get_auto_ckpt_filename()
                    cache = Path(self.ckpt_file)
                    if not cache.is_file():
                        self.mw.signals.showWaitDialog.emit("Please wait for yolox model to download")
                        try:
                            cache.parent.mkdir(parents=True, exist_ok=True)
                            id = f'{self.model_name}.pth'
                            link = "https://github.com/Megvii-BaseDetection/YOLOX/releases/download/0.1.1rc0/" + id
                            response = requests.get(link, allow_redirects=True, timeout=(10, 120))
                            if not response:
                                raise RuntimeError(f'Error downloading {link}: {response.status_code}')
                            with open(self.ckpt_file, 'wb') as content:
                                content.write(response.content)
                            if os.path.isfile(self.ckpt_file):

                                hashes = {
                                    "yolox_tiny.pth": "9de513de589ac98bb92d3bca53b5af7b9acfa9b0bacb831f7999d0f7afaee8f0",
                                    "yolox_s.pth": "f55ded7181e1b0c13285c56e7790b8f0e8f8db590fe4edb37f0b7f345c913a30",
                                    "yolox_m.pth": "60076992b32da82951c90cfa7bd6ab70eba9eda243e08b940a396f60ac2d19b6",
                                    "yolox_l.pth": "1e6b7fa6240375370b2a8a8eab9066b3cdd43fd1d0bfa8d2027fd3a51def2917",
                                    "yolox_x.pth": "5652330b6ae860043f091b8f550a60c10e1129f416edfdb65c259be6caf355cf"
                                }

                                verified = False
                                if hash := self.mw.calculate_sha256(self.ckpt_file):
                                    if hash == hashes.get(id, None):
                                        verified = True
                                if not verified:
                                    os.remove(self.ckpt_file)
                                    raise RuntimeError(f'Error verifying {self.ckpt_file}')

                                logger.debug(f'model {self.ckpt_file} was downloaded succesfully')
                        except Exception as ex:
                            logger.error(f'YOLO model {link} download failure: {ex}')
                        self.mw.signals.hideWaitDialog.emit()
                else:
                    self.ckpt_file = self.mw.videoConfigure.txtFilename.text()

                self.model.load_state_dict(torch.load(self.ckpt_file, map_location="cpu")["model"])
                self.model(initializer_data.to(self.torch_device))

            if self.api == "OpenVINO" and sys.platform != "darwin":
                if not ov_model:
                    ov_model = ov.convert_model(self.model, example_input=initializer_data)
                    ov.save_model(ov_model, self.get_ov_model_filename())

                self.ov_device = self.mw.videoConfigure.cmbDevice.currentText()
                if self.ov_device != "CPU":
                    ov_model.reshape({0: [1, 3, self.res, self.res]})
                ov_config = {}
                if "GPU" in self.ov_device or ("AUTO" in self.ov_device and "GPU" in ov.Core().available_devices):
                    ov_config = {"GPU_DISABLE_WINOGRAD_CONVOLUTION": "YES"}

                self.compiled_model = ov.compile_model(ov_model, self.ov_device, ov_config)
                self.compiled_model(initializer_data)

            if not self.torch_device:
                self.torch_device = torch.device("cpu")

        except Exception as ex:
            logger.exception(MODULE_NAME + " initialization failure")
            self.mw.signals.error.emit(f'{MODULE_NAME} initialization failure - {ex}')

        self.mw.glWidget.model_loading = False

    def preprocess(self, frame):
        h = frame.shape[0]
        w = frame.shape[1]

        test_size = (self.res, self.res)
        ratio = min(test_size[0] / h, test_size[1] / w)
        inf_shape = (int(h * ratio), int(w * ratio))
        bottom = test_size[0] - inf_shape[0]
        side = test_size[1] - inf_shape[1]
        pad = (0, 0, side, bottom)

        timg = functional.to_tensor(frame).to(self.torch_device)
        timg *= 255
        timg = functional.resize(timg, inf_shape)
        timg = functional.pad(timg, pad, 114)
        timg = timg.unsqueeze(0)
        return timg
    
    def postprocess(self, outputs, confthre, size, targets):
        nmsthre = 0.65
        if isinstance(outputs, np.ndarray):
            outputs = torch.from_numpy(outputs)
        outputs = cayenue.panels.video.modules.yolox.utils.postprocess(outputs, self.num_classes, confthre, nmsthre)
        output = None
        boxes = []
        test_size = (self.res, self.res)

        h = size[0]
        w = size[1]
        ratio = min(test_size[0] / h, test_size[1] / w)

        if outputs[0] is not None:
            output = outputs[0].cpu().numpy().astype(float)
            output[:, 0:4] /= ratio
            output[:, 4] *= output[:, 5]
            output = np.delete(output, 5, 1)

            labels = output[:, 5].astype(int)
            for i in range(len(labels)):
                if labels[i] in targets:
                    boxes.append(output[i, 0:4])

        return boxes

    def __call__(self, frame, uri):
        try:
            if len(IMPORT_ERROR) or self.mw.glWidget.model_loading or not self.mw.videoConfigure:
                return
            
            if frame is None:
                return
            
            if self.mw.videoConfigure.name != MODULE_NAME:
                self.mw.videoConfigure.selTargets.barLevel.setLevel(0)
                self.mw.videoConfigure.selTargets.indAlarm.setState(0)
                return
            
            camera = self.mw.cameraPanel.getCamera(uri)
            if not camera:
                return
            if camera.videoModelSettings.module_name != MODULE_NAME:
                camera.videoModelSettings = VideoModelSettings(self.mw, camera, MODULE_NAME)
            
            if self.parameters_changed():
                self.__init__(self.mw)
                return
            
            h = frame.shape[0]
            w = frame.shape[1]

            if self.api == "OpenVINO" and self.compiled_model:
                timg = self.preprocess(frame)                
                infer_request = self.compiled_model.create_infer_request()
                infer_request.set_input_tensor(ov.Tensor(np.asarray(timg)))
                infer_request.infer()
                outputs = infer_request.get_output_tensor(0).data
            if self.api == "PyTorch" and self.model:
                timg = self.preprocess(frame)
                with torch.no_grad():
                    outputs = self.model(timg)

            return self.postprocess(outputs, camera.videoModelSettings.confidence / 100, [h, w], camera.videoModelSettings.targets)

        except Exception as ex:
            if self.last_ex != str(ex) and self.mw.videoConfigure.name == MODULE_NAME:
                logger.exception(f'{MODULE_NAME} runtime error - {ex}')
            self.last_ex = str(ex)

    def parameters_changed(self):
        result = False

        api = self.api == self.mw.videoConfigure.cmbAPI.currentText()
        name = self.model_name == self.mw.videoConfigure.cmbModelName.currentText()
        res = str(self.res) == self.mw.videoConfigure.cmbRes.currentText()

        dev = False
        if self.api == "PyTorch":
            dev = True
            if self.mw.videoConfigure.cmbDevice.currentText() != "auto":
                dev = self.torch_device_name == self.mw.videoConfigure.cmbDevice.currentText()
        if self.api == "OpenVINO":
            dev = self.ov_device == self.mw.videoConfigure.cmbDevice.currentText()

        if not api or not name or not res or not dev:
            result = True

        return result

    def get_ov_model_filename(self):
        model_name = self.mw.videoConfigure.cmbModelName.currentText()
        openvino_device = self.mw.videoConfigure.cmbDevice.currentText()
        path = os.path.join(self.mw.getCacheLocation(), "checkpoints", model_name, openvino_device, "model.xml")
        return path

    def get_auto_ckpt_filename(self):
        model_name = self.mw.videoConfigure.cmbModelName.currentText()
        path = os.path.join(self.mw.getCacheLocation(), "checkpoints", f'{model_name}.pth')
        return path

    def get_model(self, num_classes, depth, width, act):
        def init_yolo(M):
            for m in M.modules():
                if isinstance(m, nn.BatchNorm2d):
                    m.eps = 1e-3
                    m.momentum = 0.03

        in_channels = [256, 512, 1024]
        backbone = YOLOPAFPN(depth, width, in_channels=in_channels)
        head = YOLOXHead(num_classes, width, in_channels=in_channels)
        model = YOLOX(backbone, head)

        model.apply(init_yolo)
        model.head.initialize_biases(1e-2)
        return model



#/********************************************************************
# Cayenue/cayenue/panels/audio/audiopanel.py 
#
# Copyright (c) 2023  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

import os
from PyQt6.QtWidgets import QGridLayout, QWidget, \
    QLabel, QComboBox, QVBoxLayout
from PyQt6.QtCore import Qt

class AudioPanel(QWidget):
    def __init__(self, mw):
        super().__init__()
        self.mw = mw
        self.panel = None
        self.layout = QVBoxLayout(self)
        self.workerKey = "AudioPanel/worker"
        self.cmbWorkerConnected = True

        self.stdLocation = mw.getLocation() + "/cayenue/panels/audio/modules"

        self.cmbWorker = QComboBox()
        self.fillModules()
        self.cmbWorker.setCurrentText(mw.settings.value(self.workerKey, "sample.py"))
        self.cmbWorker.currentTextChanged.connect(self.cmbWorkerChanged)
        lblWorkers = QLabel("Audio Analyzer")

        self.lblCamera = QLabel("Please select a camera to enable this panel")
        self.lblMessage = QLabel()

        fixedPanel = QWidget()
        lytFixed = QGridLayout(fixedPanel)
        lytFixed.addWidget(lblWorkers,         1, 0, 1, 1)
        lytFixed.addWidget(self.cmbWorker,     1, 1, 1, 1)
        lytFixed.addWidget(QLabel(),           2, 0, 1, 1)
        lytFixed.addWidget(self.lblCamera,     3, 0, 1, 2, Qt.AlignmentFlag.AlignCenter)
        lytFixed.addWidget(self.lblMessage,    4, 0, 1, 2)
        lytFixed.setColumnStretch(1, 10)
        self.layout.addWidget(fixedPanel)

    def showEvent(self, event):
        self.lblCamera.setFocus()
        super().showEvent(event)

    def fillModules(self):
        d = self.stdLocation
        workers = [f for f in os.listdir(d) if os.path.isfile(os.path.join(d, f))]
        for worker in workers:
            if not worker.endswith(".py") or worker == "__init__.py":
                workers.remove(worker)
        workers.sort()
        self.cmbWorker.clear()
        self.cmbWorker.addItems(workers)

    def setPanel(self, panel):
        if self.panel is not None:
            self.layout.removeWidget(self.panel)
        self.panel = panel
        self.layout.addWidget(panel)
        self.layout.setStretch(1, 10)

    def cmbWorkerChanged(self, worker):
        self.mw.settings.setValue(self.workerKey, worker)
        self.mw.audioFirstPass = True
        self.mw.audioRuntimes.clear()
        self.mw.loadAudioConfigure(worker)
        self.mw.loadAudioWorker(worker)



#/********************************************************************
# Cayenue/cayenue/panels/file/controlpanel.py 
#
# Copyright (c) 2025  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

import os
from PyQt6.QtWidgets import QPushButton, QGridLayout, \
    QWidget, QSlider, QFileDialog, QMessageBox
from PyQt6.QtCore import Qt, QStandardPaths
import sys
import platform
from datetime import datetime
from loguru import logger
from pathlib import Path
from time import sleep
from .searchdialog import FileSearchDialog

class FileControlPanel(QWidget):
    def __init__(self, mw, panel):
        super().__init__()
        self.mw = mw
        self.panel = panel
        self.hideCameraKey = "filePanel/hideCameraPanel"

        self.dlgSearch = FileSearchDialog(self.mw)

        self.btnSearch = QPushButton()
        self.btnSearch.setStyleSheet(self.getButtonStyle("search"))
        self.btnSearch.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.btnSearch.clicked.connect(self.btnSearchClicked)

        self.btnSnapshot = QPushButton()
        self.btnSnapshot.setStyleSheet(self.getButtonStyle("snapshot"))
        self.btnSnapshot.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.btnSnapshot.clicked.connect(self.btnSnapshotClicked)

        self.btnRefresh = QPushButton()
        self.btnRefresh.setStyleSheet(self.getButtonStyle("refresh"))
        self.btnRefresh.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.btnRefresh.clicked.connect(self.btnRefreshClicked)

        self.btnPlay = QPushButton()
        self.btnPlay.setStyleSheet(self.getButtonStyle("play"))
        self.btnPlay.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.btnPlay.clicked.connect(self.btnPlayClicked)
        
        self.btnStop = QPushButton()
        self.btnStop.setStyleSheet(self.getButtonStyle("stop"))
        self.btnStop.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.btnStop.clicked.connect(self.btnStopClicked)

        self.btnPrevious = QPushButton()
        self.btnPrevious.setStyleSheet(self.getButtonStyle("previous"))
        self.btnPrevious.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.btnPrevious.clicked.connect(self.btnPreviousClicked)

        self.btnNext = QPushButton()
        self.btnNext.setStyleSheet(self.getButtonStyle("next"))
        self.btnNext.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.btnNext.clicked.connect(self.btnNextClicked)

        self.btnMute = QPushButton()
        self.btnMute.setStyleSheet(self.getButtonStyle("mute"))
        self.btnMute.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.btnMute.clicked.connect(self.btnMuteClicked)

        self.sldVolume = QSlider(Qt.Orientation.Horizontal)
        self.sldVolume.setValue(80)
        self.sldVolume.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.sldVolume.valueChanged.connect(self.sldVolumeChanged)

        lytMain =  QGridLayout(self)
        lytMain.addWidget(self.btnSearch,       0, 0, 1, 1)
        lytMain.addWidget(self.btnRefresh,      0, 1, 1, 1)
        lytMain.addWidget(self.btnSnapshot,     0, 2, 1, 1)
        lytMain.addWidget(self.btnPrevious,     1, 0, 1, 1)
        lytMain.addWidget(self.btnPlay,         1, 1, 1, 1)
        lytMain.addWidget(self.btnNext,         1, 2, 1, 1)
        lytMain.addWidget(self.btnStop,         1, 3, 1, 1)
        lytMain.addWidget(self.btnMute,         1, 4, 1, 1)
        lytMain.addWidget(self.sldVolume,       1, 5, 1, 1)
        lytMain.setColumnStretch(5, 10)
        lytMain.setContentsMargins(0, 0, 0, 0)

    def btnStopClicked(self):
        self.mw.closeAnyPlayingFiles()
        self.setBtnPlay()

    def btnPlayClicked(self):
        if type(self.panel).__name__ == "FilePanel":
            self.startPlayer()
        else:
            self.mw.picturePanel.playVideo()

    def startPlayer(self, file_start_from_seek=-1.0):
        uri = self.mw.filePanel.getCurrentFileURI()
        if player := self.mw.pm.getPlayer(uri):
            player.togglePaused()
        else:
            self.mw.closeAnyPlayingFiles()
            self.mw.playMedia(uri, file_start_from_seek=file_start_from_seek)
            if uri: self.mw.glWidget.focused_uri = uri

        self.setBtnPlay()

    def setBtnPlay(self):
        self.btnPlay.setStyleSheet(self.getButtonStyle("play"))
        if player := self.mw.pm.getPlayer(self.mw.filePanel.getCurrentFileURI()):
            if not player.isPaused():
                self.btnPlay.setStyleSheet(self.getButtonStyle("pause"))

    def btnMuteClicked(self):
        if player := self.mw.pm.getPlayer(self.mw.filePanel.getCurrentFileURI()):
            player.setMute(not player.isMuted())
            self.mw.filePanel.setMute(player.isMuted())
        else:
            self.mw.filePanel.setMute(not self.mw.filePanel.getMute())
        self.setBtnMute()

    def setBtnMute(self):
        self.btnMute.setStyleSheet(self.getButtonStyle("mute"))
        if not self.mw.filePanel.getMute():
            self.btnMute.setStyleSheet(self.getButtonStyle("audio"))

    def btnPreviousClicked(self):
        if type(self.panel).__name__ == "FilePanel":
            tree = self.mw.filePanel.tree
        else:
            tree = self.mw.picturePanel.tree
        index = tree.currentIndex()
        if index.isValid():
            prevIndex = tree.indexAbove(index)
            if prevIndex.isValid():
                tree.setCurrentIndex(prevIndex)
                tree.scrollTo(prevIndex)
                if type(self.panel).__name__ == "FilePanel":
                    self.mw.closeAnyPlayingFiles()                
                    if tree.model().fileInfo(prevIndex).isFile():
                        uri = tree.model().fileInfo(prevIndex).filePath()
                        self.mw.playMedia(uri)
                        self.mw.glWidget.focused_uri = uri

    def btnNextClicked(self):
        if type(self.panel).__name__ == "FilePanel":
            tree = self.mw.filePanel.tree
        else:
            tree = self.mw.picturePanel.tree
        index = tree.currentIndex()
        if index.isValid():
            fileInfo = tree.model().fileInfo(index)
            if fileInfo.isDir():
                if not tree.isExpanded(index):
                    tree.expand(index)
                    return
                
            nextIndex = tree.indexBelow(index)
            if nextIndex.isValid():
                tree.setCurrentIndex(nextIndex)
                tree.scrollTo(nextIndex)
                if type(self.panel).__name__ == "FilePanel":
                    self.mw.closeAnyPlayingFiles()
                    if tree.model().fileInfo(nextIndex).isFile():
                        uri = tree.model().fileInfo(nextIndex).filePath()
                        if uri:
                            self.mw.playMedia(uri)
                            self.mw.glWidget.focused_uri = uri

    def btnSearchClicked(self):
        camera_names = []
        path = self.mw.filePanel.dirArchive.txtDirectory.text()
        for name in os.listdir(path):
            if os.path.isdir(os.path.join(path, name)):
                valid = True
                if sys.platform == "win32" and name == "Captures":
                    valid = False
                if sys.platform == "darwin" and name == "TV":
                    valid = False
                if valid:
                    camera_names.append(name)
        camera_names = sorted(camera_names, key=lambda s: s.casefold())
        self.dlgSearch.cameras.clear()
        self.dlgSearch.cameras.addItems(camera_names)

        if not self.dlgSearch.positionInitialized:
            w = 240
            h = 320
            x = int(self.mw.x() + self.mw.width()/2 - w/2)
            y = int(self.mw.y() + self.mw.height()/2 - h/2)
            self.dlgSearch.move(x, y)
        self.dlgSearch.show()

    def btnRefreshClicked(self):
        self.panel.refresh()

    def btnSnapshotClicked(self):
        try:
            if player := self.mw.filePanel.getCurrentlyPlayingFile():
                picture_dir = QStandardPaths.standardLocations(QStandardPaths.StandardLocation.PicturesLocation)[0]
                filename = '{0:%Y%m%d%H%M%S.jpg}'.format(datetime.now())
                filepath = Path(picture_dir) / filename
                if platform.system() == "Linux":
                    filename = QFileDialog.getSaveFileName(self, "Save File As", str(filepath), options=QFileDialog.Option.DontUseNativeDialog)[0]
                else:
                    filename = QFileDialog.getSaveFileName(self, "Save File As", str(filepath))[0]

                answer = QMessageBox.StandardButton.Yes
                if not filename.endswith(".jpg"):
                    filename += ".jpg"
                    if Path(filename).is_file():
                        answer = QMessageBox.question(self.mw, "File Exists", "You are about to overwrite an existing file, are you sure you wnat to do this?")
                if answer == QMessageBox.StandardButton.Yes:
                    player.image.save(filename)
                    logger.debug(f'Snapshot saved as {filename}')
        except Exception as ex:
            logger.error(f"ControlPanel btnSnapshotClicked exception: {ex}")

    def sldVolumeChanged(self, value):
        self.mw.filePanel.setVolume(value)
        if player := self.mw.filePanel.getCurrentlyPlayingFile():
            player.setVolume(value)

    def setSldVolume(self):
        self.sldVolume.setValue(self.mw.filePanel.getVolume())

    def getButtonStyle(self, name):
        strStyle = "QPushButton { image : url(image:%1.png); } QPushButton:hover { image : url(image:%1_hi.png); } QPushButton:pressed { image : url(image:%1_lo.png); }"
        strStyle = strStyle.replace("%1", name)
        return strStyle



#/********************************************************************
# Cayenue/cayenue/panels/settings/discover.py 
#
# Copyright (c) 2024  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

from PyQt6.QtWidgets import QLineEdit, QGridLayout, QWidget, QCheckBox, \
    QLabel, QComboBox, QPushButton, QDialog, QDialogButtonBox, \
    QRadioButton, QGroupBox
from PyQt6.QtCore import Qt, QRegularExpression
from PyQt6.QtGui import QRegularExpressionValidator
from loguru import logger
from cayenue.enums import ProxyType
import libonvif as onvif

class AddCameraDialog(QDialog):
    def __init__(self, mw):
        super().__init__(mw)
        self.mw = mw

        ipRange = "(?:[0-1]?[0-9]?[0-9]|2[0-4][0-9]|25[0-5])" 
        ipRegex = QRegularExpression("^" + ipRange + "\\." + ipRange + "\\." + ipRange + "\\." + ipRange + "$")
        ipValidator = QRegularExpressionValidator(ipRegex, self)           
    
        self.setWindowTitle("Add Camera")
        self.txtIPAddress = QLineEdit()
        self.txtIPAddress.setValidator(ipValidator)
        self.lblIPAddress = QLabel("IP Address")
        self.txtOnvifPort = QLineEdit()
        self.lblOnvifPort = QLabel("Onvif Port")

        buttonBox = QDialogButtonBox( \
            QDialogButtonBox.StandardButton.Ok | \
            QDialogButtonBox.StandardButton.Cancel)

        lytMain = QGridLayout(self)
        lytMain.addWidget(self.lblIPAddress,   1, 0, 1, 1)
        lytMain.addWidget(self.txtIPAddress,   1, 1, 1, 1)
        lytMain.addWidget(self.lblOnvifPort,   2, 0, 1, 1)
        lytMain.addWidget(self.txtOnvifPort,   2, 1, 1, 1)
        lytMain.addWidget(buttonBox,           5, 0, 1, 2)

        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)

        self.txtIPAddress.setFocus()

class DiscoverOptions(QWidget):
    def __init__(self, mw):
        super().__init__()
        self.mw = mw

        self.interfaceKey = "settings/interface"
        self.autoDiscoverKey = "settings/autoDiscover"
        self.discoveryTypeKey = "settings/discoveryType"
        self.autoStartKey = "settings/autoStart"
        self.scanAllKey = "settings/scanAll"
        self.cameraListKey = "settings/cameraList"


        self.grpDiscoverType = QGroupBox("Set Camera Discovery Method")
        self.radDiscover = QRadioButton("Discover Broadcast", self.grpDiscoverType )
        self.radDiscover.setChecked(int(self.mw.settings.value(self.discoveryTypeKey, 1)))
        self.radDiscover.toggled.connect(self.radDiscoverToggled)
        self.radCached = QRadioButton("Cached Addresses", self.grpDiscoverType )
        self.radCached.setChecked(not self.radDiscover.isChecked())
        lytDiscoverType = QGridLayout(self.grpDiscoverType )
        lytDiscoverType.addWidget(self.radDiscover,   0, 0, 1, 1)
        lytDiscoverType.addWidget(self.radCached,     0, 1, 1, 1)

        self.chkScanAllNetworks = QCheckBox("Scan All Networks During Discovery")
        self.chkScanAllNetworks.setChecked(int(mw.settings.value(self.scanAllKey, 1)))
        self.chkScanAllNetworks.stateChanged.connect(self.scanAllNetworksChecked)
        self.cmbInterfaces = QComboBox()
        intf = self.mw.settings.value(self.interfaceKey, "")
        self.lblInterfaces = QLabel("Network")
        session = onvif.Session()
        session.getActiveInterfaces()
        i = 0
        while len(session.active_interface(i)) > 0 and i < 16:
            if session.active_interface(i) != "172.17.0.1":
                self.cmbInterfaces.addItem(session.active_interface(i))
            i += 1
        if len(intf) > 0:
            self.cmbInterfaces.setCurrentText(intf)
        self.cmbInterfaces.currentTextChanged.connect(self.cmbInterfacesChanged)
        self.cmbInterfaces.setEnabled(not self.chkScanAllNetworks.isChecked())
        self.lblInterfaces.setEnabled(not self.chkScanAllNetworks.isChecked())

        self.btnAddCamera = QPushButton("Add Camera")
        self.btnAddCamera.clicked.connect(self.btnAddCameraClicked)

        self.chkAutoDiscover = QCheckBox("Auto Discovery")
        self.chkAutoDiscover.setChecked(bool(int(mw.settings.value(self.autoDiscoverKey, 0))))
        self.chkAutoDiscover.stateChanged.connect(self.autoDiscoverChecked)
        
        self.chkAutoStart = QCheckBox("Auto Start")
        self.chkAutoStart.setChecked(bool(int(mw.settings.value(self.autoStartKey, 0))))
        self.chkAutoStart.stateChanged.connect(self.autoStartChecked)

        pnlInterface = QGroupBox("Discovery Options")
        lytInterface = QGridLayout(pnlInterface)
        lytInterface.addWidget(self.grpDiscoverType,     0, 0, 1, 2)
        lytInterface.addWidget(self.chkScanAllNetworks,  2, 0, 1, 2)
        lytInterface.addWidget(self.lblInterfaces,       4, 0, 1, 1)
        lytInterface.addWidget(self.cmbInterfaces,       4, 1, 1, 1)
        lytInterface.addWidget(self.btnAddCamera,        5, 0, 1, 2, Qt.AlignmentFlag.AlignCenter)
        lytInterface.setColumnStretch(1, 10)
        lytInterface.setContentsMargins(10, 10, 10, 10)

        lytMain = QGridLayout(self)
        lytMain.addWidget(pnlInterface,          0, 0, 1, 2)
        lytMain.addWidget(QLabel(),              1, 0, 1, 2)
        lytMain.addWidget(self.chkAutoDiscover,  2, 0, 1, 1)
        lytMain.addWidget(self.chkAutoStart,     2, 1, 1, 1)
        lytMain.addWidget(QLabel(),              3, 0, 1, 2)
        lytMain.setRowStretch(3, 10)

        self.radDiscoverToggled(self.radDiscover.isChecked())

    def radDiscoverToggled(self, checked):
        self.chkScanAllNetworks.setEnabled(checked)
        if self.chkScanAllNetworks.isChecked():
            self.lblInterfaces.setEnabled(False)
            self.cmbInterfaces.setEnabled(False)
        else:
            self.lblInterfaces.setEnabled(checked)
            self.cmbInterfaces.setEnabled(checked)
        self.mw.settings.setValue(self.discoveryTypeKey, int(checked))

    def scanAllNetworksChecked(self, state):
        self.mw.settings.setValue(self.scanAllKey, state)
        self.cmbInterfaces.setEnabled(not self.chkScanAllNetworks.isChecked())
        self.lblInterfaces.setEnabled(not self.chkScanAllNetworks.isChecked())

    def cmbInterfacesChanged(self, network):
        self.mw.settings.setValue(self.interfaceKey, network)

    def btnAddCameraClicked(self):
        dlg = AddCameraDialog(self.mw)
        if dlg.exec():
            ip_address = dlg.txtIPAddress.text()
            onvif_port = dlg.txtOnvifPort.text()
            if not len(onvif_port):
                onvif_port = "80"
            xaddrs = f'http://{ip_address}:{onvif_port}/onvif/device_service'
            logger.debug(f'Attempting to add camera manually using xaddrs: {xaddrs}')
            data = onvif.Data()
            data.errorCallback = self.mw.cameraPanel.errorCallback
            data.infoCallback = self.mw.cameraPanel.infoCallback
            data.setSetting = self.mw.settings.setValue
            data.getSetting = self.mw.settings.value
            data.getData = self.mw.cameraPanel.getData
            data.getCredential = self.mw.cameraPanel.getCredential
            data.setXAddrs(xaddrs)
            data.alias = ip_address
            data.manual_fill()
            if self.mw.settingsPanel.proxy == ProxyType.SERVER:
                self.mw.settingsPanel.proxy.setMediaMTXProxies()

    def autoDiscoverChecked(self, state):
        self.mw.settings.setValue(self.autoDiscoverKey, state)

    def autoStartChecked(self, state):
        self.mw.settings.setValue(self.autoStartKey, state)

